#include <list> // list, iterator
#include <vector>
#include <time.h> // FPS
#include <glut.h>

#include <math.h> // M_PI, pow, sqrt, sin, cos

#if !defined( M_PI )		/* Windows nao tem definicao de M_PI */
#define M_PI 		3.14159265358979323846
#endif

#define FRAME_RATE_SAMPLES 60
#define MODO_TRANSFORMACAO_CAMERA 1
#define MODO_TRANSFORMACAO_OBJETOS 2
#define TIPO_TRANSFORMACAO_TRANSLACAO 3
#define TIPO_TRANSFORMACAO_ESCALA 4
#define TIPO_TRANSFORMACAO_ROTACAO 5
#define MODO_TRANSFORMACAO_LUZ 6

using namespace std; // espaco de nomes padrao

int Window_ID; // identificador da janela
int tipoTransformacao = 3; // tipo de transformacao
int modoTransformacao = 1; // modo de transformacao (camera ou objetos)
int modoLuz = 6;//mudança luz
int FrameCount = 0; // contagem de frames
double FrameRate = 0; // quadros por segundo
bool selecionando = false; // marca se esta aguardando a selecao de algum objeto
int objetoSelecionado = -1; // ultimo objeto seleciona (-1 = nenhum objeto selecionado)
char selecaoObjeto[30] = "Objeto Selecionado: nenhum"; // texto informativo sobre qual objeto esta selecionado
int Win_Width = 640; // largura da janela
int Win_Height = 480; // altura da janela

bool luzSelecionada = false; //marca se esta aguardando a seleçao de alguma luz
int LuzSelecao = -1; // seleciona ultima luz(-1 = nenhuma luz selecionada)
char selecaoLuz[30] = "Luz Selecionada: nenhuma"; // texto informativo sobre qual luz esta selecionada

// metodo para selecao dos modos e tipos de transformacoes
// recebe como parametro um numero de 1 a 5, sendo:
// 1 - transformacoes serao aplicadas na camera
// 2 - transformacoes serao aplicadas nos objetos
// 3 - transformacao a ser realizada sera a translacao
// 4 - transformacao a ser realizada sera a escala
// 5 - transformacao a ser realizada sera a rotacao
// 6 - transformacoes serao aplicadas nas luzes
void selecionaModo(int opcao)
{	
	if(opcao > 6) return;
	if(opcao < 3) modoTransformacao = opcao;
	if(opcao > 5) modoLuz = opcao;
	else tipoTransformacao = opcao;
}

/*! metodo para ativar a projecao paralela ortogonal */
void setOrthographicProjection()
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
		glLoadIdentity();
		gluOrtho2D(0, Win_Width, 0, Win_Height); // set a 2D orthographic projection
		glScalef(1, -1, 1); // inverte o eixo y, para baixo eh positivo
		glTranslatef(0, -Win_Height, 0); // move a origem do conta inferior esquerdo para o conto superior esquerdo
		glMatrixMode(GL_MODELVIEW);
}

/*! metodo para voltar da projecao paralela ortogonal para a projecao perspectiva */
void resetPerspectiveProjection()
{
		glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

/*! metodo para desenhar um texto os parametros sao:
 x, y, z: posicao do texto
 font: tipo da fonte
 string: sequencia de caracteres a ser desenhado */
void renderBitmapString(float x, float y, float z, void *font, char *string)
{  
	char *c;
	glRasterPos3f(x, y, z);
	for(c=string; *c != '\0'; c++) glutBitmapCharacter(font, *c);
}

/*! coleta informacoes sobre a quantidade de quadros por segundo */
void FPS()
{
	static clock_t last = 0;
	clock_t now;
	double delta;

	if(++FrameCount >= FRAME_RATE_SAMPLES)
	{
		now = clock();
		delta = (now - last) / (double)CLOCKS_PER_SEC;
		last = now;
		FrameRate = FRAME_RATE_SAMPLES / delta;
		FrameCount = 0;
	}
}

/*! desenha os textos 2D que ficam fixos na janela */
void draw2DText()
{
	char buf[50];
	setOrthographicProjection();
	glPushMatrix();
		glLoadIdentity();
		glColor4f(0.2, 0.3, 0.9, 0.75);
		renderBitmapString(15, 30, 0, GLUT_BITMAP_TIMES_ROMAN_24, "Projeto de CG - Fases 3 e 4");
		sprintf(buf,"FPS: %f Frame: %2d", FrameRate, FrameCount);
		renderBitmapString(15, 50, 0, GLUT_BITMAP_HELVETICA_12, buf);
		if(modoTransformacao == MODO_TRANSFORMACAO_CAMERA) sprintf(buf,"Modo de Transformacao: Camera");
		if(modoTransformacao == MODO_TRANSFORMACAO_OBJETOS) sprintf(buf,"Modo de Transformacao: Objetos");
		renderBitmapString(15, Win_Height-50, 0, GLUT_BITMAP_HELVETICA_10, buf);
		if(tipoTransformacao == TIPO_TRANSFORMACAO_TRANSLACAO) sprintf(buf,"Tipo de Transformacao: Translacao");
		if(tipoTransformacao == TIPO_TRANSFORMACAO_ESCALA) sprintf(buf,"Tipo de Transformacao: Escala");
		if(tipoTransformacao == TIPO_TRANSFORMACAO_ROTACAO) sprintf(buf,"Tipo de Transformacao: Rotacao");
		renderBitmapString(15, Win_Height-40, 0, GLUT_BITMAP_HELVETICA_10, buf);
		if(modoLuz == MODO_TRANSFORMACAO_LUZ) sprintf(buf,"Luz : ");
		renderBitmapString(15, Win_Height-20, 0, GLUT_BITMAP_HELVETICA_10, buf);
		if(objetoSelecionado >= 0)
		{
			if(selecionando) sprintf(buf,"%s", selecaoObjeto);
			else sprintf(buf,"%s: %d", selecaoObjeto, objetoSelecionado + 1);
		}
		else sprintf(buf,"%s", selecaoObjeto);
		renderBitmapString(15, Win_Height-30, 0, GLUT_BITMAP_HELVETICA_10, buf);
	glPopMatrix();
	resetPerspectiveProjection();
}

class Camera
{
private:
	double x0, y0, z0; // posicao da camera
	double xf, yf, zf; // ponto focal
	double UPx, UPy, UPz; // vetor up para orientacao do plano de visualizacao da camera
	double nearest; // mais proximo que a camera consegue enxergar
	double farest; // mais longe que a camera consegue enxergar
	double fov; // field-of-view
	double aspect_ratio; // w/h
	
public:
	Camera()
	{
		x0 = 0;
		y0 = 0;
		z0 = 10;
		xf = yf = zf = 0;
		UPx = 0;
		UPy = 1;
		UPz = 0;
		nearest = 0.0001;
		farest = 500;
		fov = 60;
		aspect_ratio = 1.34;
	}
	
	~Camera()
	{
		// wahzap?
	}
	
	// setters
	
	void setPosicao(double x0, double y0, double z0)
	{
		this -> x0 = x0;
		this -> y0 = y0;
		this -> z0 = z0;
	}
	
	void setPontoFocal(double xf, double yf, double zf)
	{
		this -> xf = xf;
		this -> yf = yf;
		this -> zf = zf;
	}
	
	void setVetorUP(double UPx, double UPy, double UPz)
	{
		this -> UPx = UPx;
		this -> UPy = UPy;
		this -> UPz = UPz;
	}
	
	void setNear(double nearest) { this -> nearest = nearest; }
	void setFar(double farest) { this -> farest = farest; }
	void setFov(double fov) { this -> fov = fov; }
	void setAspect(double aspect_ratio) { this -> aspect_ratio = aspect_ratio; }
	
	// getters
	
	double getPosicaoX() { return x0; }
	double getPosicaoY() { return y0; }
	double getPosicaoZ() { return z0; }
	double getPontoFocalX() { return xf; }
	double getPontoFocalY() { return yf; }
	double getPontoFocalZ() { return zf; }
	double getVetorUPX() { return UPx; }
	double getVetorUPY() { return UPy; }
	double getVetorUPZ() { return UPz; }
	double getNear() { return nearest; }
	double getFar() { return farest; }
	double getFov() { return fov; }
	double getAspect() { return aspect_ratio; }
};

/*!
	Abstrao de Material do OpenGL
*/
class Material
{
private:
   	// Cor do objeto em RGBA (Diffuse)
   	float cor[4];
	float specular[4]; // Cor de 'highlight' (Specular)
	float brilho; // Intensidade do brilho

public:

	Material();
	Material(float[4],float[4],float);
	virtual ~Material() { }

	void setCor(const float[4]);
	void getCor(float[4]) const;
	void setSpecular(const float[4]);
	void getSpecular(float[4]) const;
	void setBrilho(const float d ) { brilho = d; }
	float getBrilho() const { 	return brilho;	}

	Material& operator = (const Material&);

	virtual void aplicar();
	virtual void afterDraw() {}
};

Material::Material() : brilho(50.0f)
{
	for ( int i = 0; i < 4; ++i )
		specular[i] = cor[i] = 1.0f;
}

Material::Material( float c[4], float sp[4], float b ) : brilho(b)
{
	setCor(c);
	setSpecular(sp);
}

void Material::setCor(const float d[4])
{
	for ( int i = 0; i < 4; ++i )
		cor[i] = d[i];
}

void Material::setSpecular(const float d[4])
{
	for ( int i = 0; i < 4; ++i )
		specular[i] = d[i];
}

void Material::getCor(float d[4]) const
{
 	for ( int i = 0; i < 4; ++i)
		d[i] = cor[i];
}

void Material::getSpecular(float d[4]) const
{
	for ( int i = 0 ; i < 4; ++i )
		d[i] = specular[i];
}

Material& Material::operator = (const Material &d)
{
	float p[4];
	d.getSpecular(p);
	setSpecular(p);
	setBrilho(d.getBrilho());
	return *this;
}

void Material::aplicar()
{
	glMaterialfv(GL_FRONT, GL_SPECULAR, specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, &brilho);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, cor);
}

class MaterialTextura : public Material
{
protected:

	unsigned char* texture;
	unsigned int width, height;
	int id;

public:

   	MaterialTextura ( unsigned int, unsigned int );
	virtual void aplicar();
	virtual void afterDraw();
	virtual void makeTexture(char*);
};

class MaterialMarmore : public MaterialTextura
{
protected:
	void generarRuido();
	float lerpNoise( float x, float y, float z );
	float turbulence( float x, float y, float z, float scale );
	float generarCorMarmore( float x, float y, float z );
public:
    virtual void makeTexture( char* fileName = 0 );
};

/*!
	Object, primitiva de todos os objetos da cena
*/
class Object
{
private:
	double color[4]; // cor do objeto em RGBa
	GLfloat trans[4][4]; // Matriz de transformacao do objeto
	int numID;
	Material material;	

	void multi(GLfloat m[4][4]) // multiplica a matriz dada pela matriz de transformacao do objeto
	{
		int i, j, k;
		double aux[4][4];
		
		for(i = 0; i < 4; i++)
			for(j = 0; j < 4; j++)
				aux[i][j] = trans[i][j];
			
			for(i = 0; i < 4; i++)
			{
				for(j = 0; j < 4; j++)
				{
					trans[i][j] = 0;
					for(k = 0; k < 4; k++) trans[i][j] = trans[i][j] + (aux[i][k] * m[k][j]);
				}
			}
	}
	
public:
	Object()
	{
		int i = 0;
		for(; i < 4; i++) 
			color[i] = 0.0;
		for(i = 0; i < 4; i++)
		{
			for(int j = 0; j < 4; j++)
			{
				if(i == j) trans[i][j] = 1.0;
				else trans[i][j] = 0;
			}
		}
		numID = 0;
	}
	
	Object(int numID)
	{
		int i = 0;
		for(; i < 4; i++) color[i] = 0.0;
		for(i = 0; i < 4; i++)
		{
			for(int j = 0; j < 4; j++)
			{
				if(i == j) trans[i][j] = 1.0;
				else trans[i][j] = 0;
			}
		}
		this -> numID = numID;
	}	
	
	virtual ~Object()
	{
		// blah
	}
	
	virtual void render() = 0;
	
	// setters
	
	void setColor(double* color) { for(int i = 0; i < 4; i++) this -> color[i] = color[i]; }
	void setID(int numID) { this -> numID = numID; }
	
	// getters
	
	int getID() { return numID; }
	double* getColor() { return color; }
	GLfloat* getMatrizTransformacao() { return *trans; }
	
	// transformacoes
	
	void translacao(double dx, double dy, double dz) // translada o objeto
	{	
		GLfloat aux[4][4];
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(dx, dy, dz);
		glGetFloatv(GL_MODELVIEW_MATRIX, &aux[0][0]);
		glPopMatrix();
		multi(aux);
	}
	
	void escala(double kx, double ky, double kz) // aplica no objeto a escala dada
	{
		GLfloat aux[4][4];
		glPushMatrix();
		glLoadIdentity();
		glScalef(kx, ky, kz);
		glGetFloatv(GL_MODELVIEW_MATRIX, &aux[0][0]);
		glPopMatrix();
		multi(aux);
	}
	
	void rotacaoX(double angulo) // rotaciona o objeto em torno o eixo x
	{
		GLfloat aux[4][4];
		glPushMatrix();
		glLoadIdentity();
		glRotatef(angulo, 1, 0, 0);
		glGetFloatv(GL_MODELVIEW_MATRIX, &aux[0][0]);
		glPopMatrix();
		multi(aux);
	}
	
	void rotacaoY(double angulo) // rotaciona o objeto em torno o eixo y
	{
		GLfloat aux[4][4];
		glPushMatrix();
		glLoadIdentity();
		glRotatef(angulo, 0, 1, 0);
		glGetFloatv(GL_MODELVIEW_MATRIX, &aux[0][0]);
		glPopMatrix();
		multi(aux);
	}
	
	void rotacaoZ(double angulo) // rotaciona o objeto em torno o eixo z
	{
		GLfloat aux[4][4];
		glPushMatrix();
		glLoadIdentity();
		glRotatef(angulo, 0, 0, 1);
		glGetFloatv(GL_MODELVIEW_MATRIX, &aux[0][0]);
		glPopMatrix();
		multi(aux);
	}

	void doRender()
	{
		material.aplicar();
		render();
		material.afterDraw();
	}
};

class Luz
{
private:
	unsigned int id;
	float pos[4];
	float cor[4];

	bool estado;

public:
	
	Luz( int d );
	void toggle();
	void aplicar();

	unsigned int getID() const { return id; }
	void getCor( float [4] ) const;
	void setCor( float [4] );
	void setPosicao( float [4] );
	void getPosicao( float [4] ) const;

	void translacao( double, double, double );
};



Luz::Luz( int d ) : estado( false ), id(d)
{
	memset( pos, 0, sizeof(pos) );
	memset( cor, 0, sizeof(cor) );
}

void Luz::toggle()
{
	estado = !estado;
	if ( estado )
		glEnable(id);
	else
		glDisable(id);
	aplicar();
}

/*!
	Aplica as opcoes de Luz selecionadas
*/
void Luz::aplicar()
{
	glLightfv( id, GL_POSITION, pos );
	glLightfv( id, GL_DIFFUSE,  cor );
}

void Luz::getCor( float d[4] ) const
{
	for ( int i = 0; i < 4; ++i )
		d[i] = cor[i];
}

void Luz::setCor( float d[4] )
{
	for ( int i = 0; i < 4; ++i )
		cor[i] = d[i];
}

void Luz::setPosicao( float d[4] )
{
	for ( int i = 0; i < 4; ++i )
		pos[i] = d[i];
}

void Luz::getPosicao( float d[4] ) const
{
	for ( int i = 0; i < 4; ++i )
		d[i] = pos[i];
}

void Luz::translacao( double x, double y, double z )
{
	float p[4];
	p[0] = pos[0] + x;
	p[1] = pos[1] + y;
	p[2] = pos[2] + z;
	p[3] = pos[3] + 0;
	setPosicao( p );
}

class Esfera : public Object
{
	private:
		double raio;

	public:
		Esfera() : Object() { raio = 0; }
		//Esfera(double c[4], double t[4][4], double r) : Object(c, t) { raio = r; }
		Esfera(int id, double r) : Object(id) { raio = r; }

		void render()
		{
			char buf[2];
			glPushMatrix();
				glColor4f(getColor()[0], getColor()[1], getColor()[2], getColor()[3]);
				glMultMatrixf(&getMatrizTransformacao()[0]);
				glutWireSphere(raio, 50, 50); // (raio, slices, stack)
				glColor4f(1, 0, 0, 1);
				sprintf(buf,"%d", getID());
				renderBitmapString(0.5, 0.5, 0, GLUT_BITMAP_TIMES_ROMAN_24, buf);
			glPopMatrix();
		}

		double getRaio() { return raio; }
};

class Cubo : public Object
{
	private:
		double lado;

	public:
		Cubo() : Object() { lado = 0; }
		//Cubo(double c[4], double t[4][4], double l) : Object(c, t) { lado = l; }
		Cubo(int id, double l) : Object(id) { lado = l; }

		void render()
		{
			char buf[2];
			glPushMatrix();
				glColor4f(getColor()[0], getColor()[1], getColor()[2], getColor()[3]);
				glMultMatrixf(&getMatrizTransformacao()[0]);
				glutWireCube(lado);
				glColor4f(1, 0, 0, 1);
				sprintf(buf,"%d", getID());
				renderBitmapString(0.5, 0.5, 0, GLUT_BITMAP_TIMES_ROMAN_24, buf);
			glPopMatrix();
		}

		double getLado() { return lado; }
};

class Cone : public Object
{
	private:
		double raio;
		double altura;

	public:
		Cone() : Object()
		{
			raio = 0;
			altura = 0;
		}

		/*Cone(double c[4], double t[4][4], double r, double a) : Object(c, t)
		{
			raio = r;
			altura = a;
		}*/
		
		Cone(int id, double r, double a) : Object(id)
		{
			raio = r;
			altura = a;
		}

		void render()
		{
			char buf[2];
			glPushMatrix();
				glColor4f(getColor()[0], getColor()[1], getColor()[2], getColor()[3]);
				glMultMatrixf(&getMatrizTransformacao()[0]);
				glutWireCone(raio, altura, 50, 50); // (raio, altura, slices, stack)
				glColor4f(1, 0, 0, 1);
				sprintf(buf,"%d", getID());
				renderBitmapString(0.5, 0.5, 0, GLUT_BITMAP_TIMES_ROMAN_24, buf);
			glPopMatrix();
		}

		double getRaio() { return raio; }
		double getAltura() { return altura; }
};

class Toro : public Object
{
	private:
		double raio1;
		double raio2;

	public:
		Toro() : Object()
		{
			raio1 = 0;
			raio2 = 0;
		}

		/*Toro(double c[4], double t[4][4], double r1, double r2) : Object(c, t)
		{
			raio1 = r1;
			raio2 = r2;
		}*/
		
		Toro(int id, double r1, double r2) : Object(id)
		{
			raio1 = r1;
			raio2 = r2;
		}

		void render()
		{
			char buf[2];
			glPushMatrix();
				glColor4f(getColor()[0], getColor()[1], getColor()[2], getColor()[3]);
				glMultMatrixf(&getMatrizTransformacao()[0]);
				glutWireTorus(raio1, raio2, 30, 50); // (raio_interno, raio_externo, lados, aneis)
				glColor4f(1, 0, 0, 1);
				sprintf(buf,"%d", getID());
				renderBitmapString(0.5, 0.5, 0, GLUT_BITMAP_TIMES_ROMAN_24, buf);
			glPopMatrix();
		}

		double getRaio1() { return raio1; }
		double getRaio2() { return raio2; }
};

class Cilindro : public Object
{
	private:
		double raio;
		double altura;

	public:
		Cilindro() : Object()
		{
			raio = 0;
			altura = 0;
		}

		/*Cilindro(double c[4], double t[4][4], double r, double a) : Object(c, t)
		{
			raio = r;
			altura = a;
		}*/
		
		Cilindro(int id, double r, double a) : Object(id)
		{
			raio = r;
			altura = a;
		}

		void render()
		{
			char buf[2];
			glPushMatrix();
				glColor4f(getColor()[0], getColor()[1], getColor()[2], getColor()[3]);
				glMultMatrixf(&getMatrizTransformacao()[0]);
				GLUquadricObj *q = gluNewQuadric();
				gluQuadricDrawStyle(q, GLU_LINE);
				gluQuadricNormals(q, GLU_SMOOTH);
				gluCylinder(q, raio, raio, altura, 50, 50); // (quadric, raio_base, raio_top, altura, slices, stack)
				gluDeleteQuadric(q);
				glColor4f(1, 0, 0, 1);
				sprintf(buf,"%d", getID());
				renderBitmapString(0.5, 0.5, 0, GLUT_BITMAP_TIMES_ROMAN_24, buf);
			glPopMatrix();
		}

		double getRaio() { return raio; }
		double getAltura() { return altura; }
};

class Elipsoide : public Object
{    
	private:
		double A;
		double B;
		double C;

	public:
		Elipsoide() : Object()
		{
			A = 0;
			B = 0;
			C = 0;
		}

		/*Elipsoide(double co[4], double t[4][4], double a, double b, double c) : Object(co, t)
		{
			A = a;
			B = b;
			C = c;
		}*/
		
		Elipsoide(int id, double a, double b, double c) : Object(id)
		{
			A = a;
			B = b;
			C = c;
		}

		void render()
		{
			char buf[2];
			glPushMatrix();
				glColor4f(getColor()[0], getColor()[1], getColor()[2], getColor()[3]);
				glMultMatrixf(&getMatrizTransformacao()[0]);
				glScalef(A, B, C);
				glutWireSphere(1, 50, 50);
				glColor4f(1, 0, 0, 1);
				sprintf(buf,"%d", getID());
				renderBitmapString(0.5, 0.5, 0, GLUT_BITMAP_TIMES_ROMAN_24, buf);
			glPopMatrix();
		}

		double getA() { return A; }
		double getB() { return B; }
		double getC() { return C; }
};

class Paralelepipedo : public Object
{    
	private:
		double altura;
		double largura;
		double profundidade;

	public:
		Paralelepipedo() : Object()
		{
			altura = 0;
			largura = 0;
			profundidade = 0;
		}

		/*Paralelepipedo(double c[4], double t[4][4], double a, double l, double p) : Object(c, t)
		{
			altura = a;
			largura = l;
			profundidade = p;
		}*/
		
		Paralelepipedo(int id, double a, double l, double p) : Object(id)
		{
			altura = a;
			largura = l;
			profundidade = p;
		}

		void render()
		{
			char buf[2];
			glPushMatrix();
				glColor4f(getColor()[0], getColor()[1], getColor()[2], getColor()[3]);
				glMultMatrixf(&getMatrizTransformacao()[0]);
				glScalef(profundidade, largura, altura);
				glutWireCube(1);
				glColor4f(1, 0, 0, 1);
				sprintf(buf,"%d", getID());
				renderBitmapString(0.5, 0.5, 0, GLUT_BITMAP_TIMES_ROMAN_24, buf);
			glPopMatrix();
		}

		double getAltura() { return altura; }
		double getlargura() { return largura; }
		double getProfundidade() { return profundidade; }
};

class Scene
{
	private:
		list <Object *> *objects; // lista STL de objetos
		vector< Luz* > luzes;
		Camera *cam; // camera associada à cena
		double corSelecao[4]; // cor de um objeto quando esta selecionado
		
		void updateCamera() // atualiza a camera para a nova posicao, olhando para o novo ponto com a nova orientacao para cima
		{
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			gluLookAt(cam -> getPosicaoX(), cam -> getPosicaoY(), cam -> getPosicaoZ(),
					cam -> getPontoFocalX(), cam -> getPontoFocalY(), cam -> getPontoFocalZ(),
					cam -> getVetorUPX(), cam -> getVetorUPY(), cam -> getVetorUPZ());
		}

	public:
		Scene()
		{
			objects = new list <Object *>;
			cam = new Camera();
			corSelecao[0] = 0.2;
			corSelecao[1] = 0.8;
			corSelecao[2] = 0.2;
			corSelecao[3] = 0.0;
		}
		
		~Scene()
		{
			if(objects != NULL)
			{
				for ( list<Object*>::iterator it = objects->begin(); it != objects->end(); ++it )
					delete *it;
				delete objects;
			}
			delete cam;
		}
		
		void insertObject(Object *obj) { objects -> push_back(obj); } // insere um objeto na lista de objetos
		
		void insertLuz( Luz* d ) { luzes.push_back( d ); }

		void render()
		{
			for(list <Object *>::iterator i = objects -> begin(); i != objects -> end(); i++)
			{
				(*i)->doRender();
			}
		}
		
		/*! retorna da lista de objetos o objeto da posicao especificada */
		Object* getObject(int index) 
		{
			int pos = 0;
			
			for(list <Object *>::iterator i = objects -> begin(); i != objects -> end(); i++, pos++)
			{
				if(pos == index) return(*i);
			}
			return NULL;
		}
		
		double* getCorSelecao() { return corSelecao; }
		
		void reshape(int width, int height)
		{
			glMatrixMode(GL_PROJECTION); // define estado de projecao dos objetos 3D->2D
			glLoadIdentity();
			glViewport(0, 0, width, height); // define janela de visualizacao (viewport)
			gluPerspective(cam -> getFov(), width/height, cam -> getNear(), cam -> getFar()); // define volume de visualizacao (fov, aspect_ratio, near, far)
			updateCamera();
		}

		/*! faz animacao dos objetos da cena */
		void anima() 
		{
			for(list <Object *>::iterator i = objects -> begin(); i != objects -> end(); i++)
			{
				(*i)-> rotacaoX(0.05);
				(*i)-> rotacaoY(0.05);
				(*i)-> rotacaoZ(0.05);
			}
		}
		
		/*! move camera no eixo x */
		void moveCameraX(int unidade) 
		{
			cam -> setPosicao(cam -> getPosicaoX()+unidade, cam -> getPosicaoY(), cam -> getPosicaoZ());
			cam -> setPontoFocal(cam -> getPontoFocalX()+unidade, cam -> getPontoFocalY(), cam -> getPontoFocalZ());
			updateCamera();
		}
		
		/*! move camera no eixo y */
		void moveCameraY(int unidade) 
		{
			cam -> setPosicao(cam -> getPosicaoX(), cam -> getPosicaoY()+unidade, cam -> getPosicaoZ());
			cam -> setPontoFocal(cam -> getPontoFocalX(), cam -> getPontoFocalY()+unidade, cam -> getPontoFocalZ());
			updateCamera();
		}
		
		/*! move camera no eixo z */
		void moveCameraZ(int unidade) 
		{
			cam -> setPosicao(cam -> getPosicaoX(), cam -> getPosicaoY(), cam -> getPosicaoZ()+unidade);
			cam -> setPontoFocal(cam -> getPontoFocalX(), cam -> getPontoFocalY(), cam -> getPontoFocalZ()+unidade);
			updateCamera();
		}
		
		/*! aplica um zoom na camera (quantidade dada em %) */
		void Zoom(double quantidade) 
		{
			cam -> setFov(cam -> getFov() * (quantidade/100));
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			gluPerspective(cam -> getFov(), cam -> getAspect(), cam -> getNear(), cam -> getFar());
		}
		
		/*! rotaciona a camera em torno do eixo x */
		void rotacionarCameraX(int sentido) 
		{
			static double angulo = 0;
			if(sentido < 0)
			{
				angulo = angulo + ((2*M_PI)/360);
				if(angulo == (2*M_PI)) angulo = 0;
			}
			else
			{
				angulo = angulo - ((2*M_PI)/360);
				if(angulo == 0) angulo = (2*M_PI);
			}
			
			double raio = sqrt(pow(cam -> getPosicaoY(), 2) + pow(cam -> getPosicaoZ(), 2));
			cam -> setPosicao(cam -> getPosicaoX(), raio*cos(angulo), raio*sin(angulo));
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			gluLookAt(cam -> getPosicaoX(), cam -> getPosicaoY(), cam -> getPosicaoZ(),
					cam -> getPontoFocalX(), cam -> getPontoFocalY(), cam -> getPontoFocalZ(),
					1, 0, 0);
		}
		
		/*! rotaciona a camera em torno do eixo y */
		void rotacionarCameraY(int sentido) 
		{
			static double angulo = 0;
			if(sentido < 0)
			{
				angulo = angulo + ((2*M_PI)/360);
				if(angulo == (2*M_PI)) angulo = 0;
			}
			else
			{
				angulo = angulo - ((2*M_PI)/360);
				if(angulo == 0) angulo = (2*M_PI);
			}
			double raio = sqrt(pow(cam -> getPosicaoX(), 2) + pow(cam -> getPosicaoZ(), 2));
			cam -> setPosicao(raio*cos(angulo), cam -> getPosicaoY(), raio*sin(angulo));
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			gluLookAt(cam -> getPosicaoX(), cam -> getPosicaoY(), cam -> getPosicaoZ(),
					cam -> getPontoFocalX(), cam -> getPontoFocalY(), cam -> getPontoFocalZ(),
					0, 1, 0);
		}
		
		/*! rotaciona a camera em torno do eixo z */
		void rotacionarCameraZ(int sentido) 
		{
			static double angulo = 0;
			if(sentido < 0)
			{
				angulo = angulo + ((2*M_PI)/360);
				if(angulo == (2*M_PI)) angulo = 0;
			}
			else
			{
				angulo = angulo - ((2*M_PI)/360);
				if(angulo == 0) angulo = (2*M_PI);
			}
			double raio = sqrt(pow(cam -> getPosicaoX(), 2) + pow(cam -> getPosicaoY(), 2));
			cam -> setPosicao(raio*cos(angulo), raio*sin(angulo), cam -> getPosicaoZ());
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			gluLookAt(cam -> getPosicaoX(), cam -> getPosicaoY(), cam -> getPosicaoZ(),
					cam -> getPontoFocalX(), cam -> getPontoFocalY(), cam -> getPontoFocalZ(),
					0, 0, 1);
		}
};

Scene *cena;

void display(void)
{	
	glClear(GL_COLOR_BUFFER_BIT); // limpa a janela
	cena -> render();
	draw2DText();
	glutSwapBuffers();
	FPS();
}

void clear()
{
	delete cena;
}

void specialKeyPressed(int key, int x, int y)
{
	switch(key)
	{
		case GLUT_KEY_PAGE_UP:
			switch(modoTransformacao)
			{
				case MODO_TRANSFORMACAO_OBJETOS:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> translacao(0, 0, 1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> escala(1, 1, 1.1);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> rotacaoZ(-1);
							break;
					}
					break;
					
				case MODO_TRANSFORMACAO_CAMERA:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							cena -> moveCameraZ(-1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							cena -> rotacionarCameraZ(1);
							break;
					}
					break;
			}
			break;

		case GLUT_KEY_PAGE_DOWN:
			switch(modoTransformacao)
			{
				case MODO_TRANSFORMACAO_OBJETOS:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> translacao(0, 0, -1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> escala(1, 1, 0.9);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> rotacaoZ(1);
							break;
					}
					break;
					
				case MODO_TRANSFORMACAO_CAMERA:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							cena -> moveCameraZ(1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							cena -> rotacionarCameraZ(-1);
							break;
					}
					break;
			}
			break;

		case GLUT_KEY_UP:
			switch(modoTransformacao)
			{
				case MODO_TRANSFORMACAO_OBJETOS:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> translacao(0, 1, 0);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> escala(1, 1.1, 1);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> rotacaoX(1);
							break;
					}
					break;
					
				case MODO_TRANSFORMACAO_CAMERA:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							cena -> moveCameraY(-1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							cena -> Zoom(90);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							cena -> rotacionarCameraX(1);
							break;
					}
					break;
			}
			break;

		case GLUT_KEY_DOWN:
			switch(modoTransformacao)
			{
				case MODO_TRANSFORMACAO_OBJETOS:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> translacao(0, -1, 0);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> escala(1, 0.9, 1);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> rotacaoX(-1);
							break;
					}
					break;
					
				case MODO_TRANSFORMACAO_CAMERA:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							cena -> moveCameraY(1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							cena -> Zoom(110);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							cena -> rotacionarCameraX(-1);
							break;
					}
					break;
			}
			break;

		case GLUT_KEY_LEFT:
			switch(modoTransformacao)
			{
				case MODO_TRANSFORMACAO_OBJETOS:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> translacao(-1, 0, 0);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> escala(0.9, 1, 1);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> rotacaoY(1);
							break;
					}
					break;
					
				case MODO_TRANSFORMACAO_CAMERA:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							cena -> moveCameraX(1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							cena -> rotacionarCameraY(1);
							break;
					}
					break;
			}
			break;

		case GLUT_KEY_RIGHT:
			switch(modoTransformacao)
			{
				case MODO_TRANSFORMACAO_OBJETOS:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> translacao(1, 0, 0);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> escala(1.1, 1, 1);
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							if(objetoSelecionado == -1) break;
							(cena -> getObject(objetoSelecionado)) -> rotacaoY(-1);
							break;
					}
					break;
					
				case MODO_TRANSFORMACAO_CAMERA:
					switch(tipoTransformacao)
					{
						case TIPO_TRANSFORMACAO_TRANSLACAO:
							cena -> moveCameraX(-1);
							break;
							
						case TIPO_TRANSFORMACAO_ESCALA:
							break;
							
						case TIPO_TRANSFORMACAO_ROTACAO:
							cena -> rotacionarCameraY(-1);
							break;
					}
					break;
			}
			break;
			
		default: break;
	}
}

void keyPressed(unsigned char key, int x, int y)
{
	double defaultColor[4] = {0, 0, 0, 0};
	Object *aux = cena -> getObject(objetoSelecionado);
	switch(key)
	{
		case '1':
			if(selecionando)
			{
				strcpy(selecaoObjeto, "Objeto Selecionado");
				if(aux != NULL) aux -> setColor(defaultColor);
				objetoSelecionado = 0;
				(cena -> getObject(0)) -> setColor(cena -> getCorSelecao());
				selecionando = false;
			}
			break;
			
		case '2':
			if(selecionando)
			{
				strcpy(selecaoObjeto, "Objeto Selecionado");
				if(aux != NULL) aux -> setColor(defaultColor);
				objetoSelecionado = 1;
				(cena -> getObject(1)) -> setColor(cena -> getCorSelecao());
				selecionando = false;
			}
			break;
			
		case '3':
			if(selecionando)
			{
				strcpy(selecaoObjeto, "Objeto Selecionado");
				if(aux != NULL) aux -> setColor(defaultColor);
				objetoSelecionado = 2;
				(cena -> getObject(2)) -> setColor(cena -> getCorSelecao());
				selecionando = false;
			}
			break;
			
		case '4':
			if(selecionando)
			{
				strcpy(selecaoObjeto, "Objeto Selecionado");
				if(aux != NULL) aux -> setColor(defaultColor);
				objetoSelecionado = 3;
				(cena -> getObject(3)) -> setColor(cena -> getCorSelecao());
				selecionando = false;
			}
			break;
			
		case '5':
			if(selecionando)
			{
				strcpy(selecaoObjeto, "Objeto Selecionado");
				if(aux != NULL) aux -> setColor(defaultColor);
				objetoSelecionado = 4;
				(cena -> getObject(4)) -> setColor(cena -> getCorSelecao());
				selecionando = false;
			}
			break;
			
		case '6':
			if(selecionando)
			{
				strcpy(selecaoObjeto, "Objeto Selecionado");
				if(aux != NULL) aux -> setColor(defaultColor);
				objetoSelecionado = 5;
				(cena -> getObject(5)) -> setColor(cena -> getCorSelecao());
				selecionando = false;
			}
			break;
			
		case '7':
			if(selecionando)
			{
				strcpy(selecaoObjeto, "Objeto Selecionado");
				if(aux != NULL) aux -> setColor(defaultColor);
				objetoSelecionado = 6;
				(cena -> getObject(6)) -> setColor(cena -> getCorSelecao());
				selecionando = false;
			}
			break;
			
		case 'o':
			selecionaModo(MODO_TRANSFORMACAO_OBJETOS);
			break;
			
		case 'c':
			selecionaModo(MODO_TRANSFORMACAO_CAMERA);
			break;
			
		case 't':
			selecionaModo(TIPO_TRANSFORMACAO_TRANSLACAO);
			break;
			
		case 'e':
			selecionaModo(TIPO_TRANSFORMACAO_ESCALA);
			break;
			
		case 'r':
			selecionaModo(TIPO_TRANSFORMACAO_ROTACAO);
			break;
			
		case 'n':
			strcpy(selecaoObjeto, "Selecione o Objeto...");
			selecionando = true;
			break;
			
		case 'w':
			glutReshapeWindow(640, 480);
			glutPositionWindow(glutGet(GLUT_SCREEN_WIDTH)/4, glutGet(GLUT_SCREEN_HEIGHT)/4);
			break;
			
		case 'f':
			glutFullScreen();
			break;
		
		case 'l':
			strcpy(selecaoLuz, "Selecione a Luz...");
			selecionando = true;
			break;
			
		case 27:
			clear();
			glutDestroyWindow(Window_ID);
			abort();
			break;
			
		default: break;
	}
}

void idle()
{
	//cena -> anima(); // animacao
	glutPostRedisplay();
}

void reshape(int width, int height)
{
	if(height == 0) height = 1;
	Win_Width = width;
	Win_Height = height;
	cena -> reshape(width, height);
}

/*!
	Inicializa os objetos da cena
*/
void init(void)
{
	glClearColor(1.0, 1.0, 1.0, 0.0); // cor para limpar a tela (branco)
	glShadeModel(GL_SMOOTH); //// modelo de iluminacao basico

	float cor[] = { 0.0,0.0,0.0,0.0 };
	float light_position[] = { 1.0, 2.0, 10.0, 1.0 };

	Luz light0(GL_LIGHT0); cor[0]=0.6;cor[1]=0.6;cor[2]=0.6; light0.setCor(cor);
	Luz light1(GL_LIGHT1); cor[0]=0.0;cor[1]=0.0;cor[2]=1.0; light1.setCor(cor);
	Luz light2(GL_LIGHT2); cor[0]=0.0;cor[1]=1.0;cor[2]=0.0; light2.setCor(cor);
	Luz light3(GL_LIGHT3); cor[0]=0.0;cor[1]=1.0;cor[2]=1.0; light3.setCor(cor);
	Luz light4(GL_LIGHT4); cor[0]=1.0;cor[1]=0.0;cor[2]=0.0; light4.setCor(cor);
	Luz light5(GL_LIGHT5); cor[0]=1.0;cor[1]=0.0;cor[2]=1.0; light5.setCor(cor);
	Luz light6(GL_LIGHT6); cor[0]=1.0;cor[1]=1.0;cor[2]=0.0; light6.setCor(cor);
	Luz light7(GL_LIGHT7); cor[0]=1.0;cor[1]=1.0;cor[2]=1.0; light7.setCor(cor);

	light0.setPosicao(light_position);
	light1.setPosicao(light_position);
	light2.setPosicao(light_position);
	light3.setPosicao(light_position);
	light4.setPosicao(light_position);
	light5.setPosicao(light_position);
	light6.setPosicao(light_position);
	light7.setPosicao(light_position);

	
	cena = new Scene(); // cria cena
	
	
	/*GLfloat cor_ambiente[] = { 1.0, 0.0, 0.0, 1.0 };
	glLightfv(GL_LIGHT0, GL_AMBIENT, cor_ambiente);
	glEnable(GL_LIGHTING); // ativa iluminacao
   	glEnable(GL_LIGHT0); // numero da fonte de luz. podem existir varias fontes de luz, GL_LIGHT1, GL_LIGHT2... normalmente ateh GL_LIGHT7
 */

	// insere luz
	cena->insertLuz(&light0);
	cena->insertLuz(&light1);
	cena->insertLuz(&light2);
	cena->insertLuz(&light3);
	cena->insertLuz(&light4);
	cena->insertLuz(&light5);
	cena->insertLuz(&light6);
	cena->insertLuz(&light7);

	
	// insere objetos na cena
	cena -> insertObject(new Paralelepipedo(1, 1, 0.5, 0.2));
	cena -> insertObject(new Cilindro(2, 0.3, 1));
	cena -> insertObject(new Elipsoide(3, 0.3, 0.8, 0.3));
	cena -> insertObject(new Esfera(4, 0.5));
	cena -> insertObject(new Cubo(5, 0.5));
	cena -> insertObject(new Cone(6, 0.5, 1.5));
	cena -> insertObject(new Toro(7, 0.1, 0.5));
	
	// posiciona objetos
	(cena -> getObject(0)) -> translacao(-3, 2, 0);
	(cena -> getObject(1)) -> translacao(0, 4, 0);
	(cena -> getObject(2)) -> translacao(3, 2, 0);
	(cena -> getObject(3)) -> translacao(-3, -2, 0);
	(cena -> getObject(4)) -> translacao(0, -4, 0);
	(cena -> getObject(5)) -> translacao(3, -2, 0);
	(cena -> getObject(6)) -> translacao(0, 0, 0);
	
}

int main(int argc, char** argv)
{	
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(Win_Width, Win_Height);
	glutInitWindowPosition(glutGet(GLUT_SCREEN_WIDTH)/4, glutGet(GLUT_SCREEN_HEIGHT)/4);
	Window_ID = glutCreateWindow("Projeto de CG - Fase 3");
	glutFullScreen();
	init();
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyPressed);
	glutSpecialFunc(specialKeyPressed);
	glutIdleFunc(idle);
	glutDisplayFunc(display);
	glutMainLoop();
	return 0;
}
